# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from .NLGElement                    import *
from ..features.ClauseStatus        import *
from ..features.DiscourseFunction   import *
from ..features.Feature             import *
from ..features.InternalFeature     import *
from .CoordinatedPhraseElement      import *
from .StringElement                 import *
from .LexicalCategory               import *

# This class defines a phrase.
class PhraseElement(NLGElement):
    def __init__(self, newCategory):
        super().__init__()
        self.setCategory(newCategory)
        self.setFeature(Feature.ELIDED, False)

    # This method retrieves the child components of this phrase.
    # @Override
    def getChildren(self):
        children = []
        category = self.getCategory()
        currentElement = None
        if isinstance(category, PhraseCategory):
            if category == PhraseCategory.CLAUSE:
                currentElement = self.getFeatureAsElement(Feature.CUE_PHRASE)
                if currentElement is not None:
                    children.append(currentElement)
                children.extend(self.getFeatureAsElementList(InternalFeature.FRONT_MODIFIERS))
                children.extend(self.getFeatureAsElementList(InternalFeature.PREMODIFIERS))
                children.extend(self.getFeatureAsElementList(InternalFeature.SUBJECTS))
                children.extend(self.getFeatureAsElementList(InternalFeature.VERB_PHRASE))
                children.extend(self.getFeatureAsElementList(InternalFeature.COMPLEMENTS))
            elif category == PhraseCategory.NOUN_PHRASE:
                currentElement = self.getFeatureAsElement(InternalFeature.SPECIFIER)
                if currentElement is not None:
                    children.append(currentElement)
                children.extend(self.getFeatureAsElementList(InternalFeature.PREMODIFIERS))
                currentElement = self.getHead()
                if currentElement is not None:
                    children.append(currentElement)
                children.extend(self.getFeatureAsElementList(InternalFeature.COMPLEMENTS))
                children.extend(self.getFeatureAsElementList(InternalFeature.POSTMODIFIERS))
            elif category == PhraseCategory.VERB_PHRASE:
                children.extend(self.getFeatureAsElementList(InternalFeature.PREMODIFIERS))
                currentElement = self.getHead()
                if currentElement is not None:
                    children.append(currentElement)
                children.extend(self.getFeatureAsElementList(InternalFeature.COMPLEMENTS))
                children.extend(self.getFeatureAsElementList(InternalFeature.POSTMODIFIERS))
            elif category == PhraseCategory.CANNED_TEXT:
                pass # Do nothing
            else:
                children.extend(self.getFeatureAsElementList(InternalFeature.PREMODIFIERS))
                currentElement = self.getHead()
                if currentElement is not None:
                    children.append(currentElement)
                children.extend(self.getFeatureAsElementList(InternalFeature.COMPLEMENTS))
                children.extend(self.getFeatureAsElementList(InternalFeature.POSTMODIFIERS))
        return children

    # Sets the head, or main component, of this current phrase.
    def setHead(self, newHead):
        if newHead is None:
            self.removeFeature(InternalFeature.HEAD)
            return
        if isinstance(newHead, NLGElement):
            headElement = newHead
        else:
            headElement = StringElement(str(newHead))
        self.setFeature(InternalFeature.HEAD, headElement)

    # Retrieves the current head of this phrase.
    def getHead(self):
        return self.getFeatureAsElement(InternalFeature.HEAD)

    # Adds a new complement to the phrase element.
    def addComplement(self, newComplement):
        complements = self.getFeatureAsElementList(InternalFeature.COMPLEMENTS)
        if complements is None:
            complements = []
        if isinstance(newComplement, NLGElement):
            # check if the new complement has a discourse function; if not, assume object
            if not newComplement.hasFeature(InternalFeature.DISCOURSE_FUNCTION):
                newComplement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.OBJECT)
            complements.append(newComplement)
            self.setFeature(InternalFeature.COMPLEMENTS, complements)
            if newComplement.isA(PhraseCategory.CLAUSE) or isinstance(newComplement, CoordinatedPhraseElement):
                newComplement.setFeature(InternalFeature.CLAUSE_STATUS, ClauseStatus.SUBORDINATE)
                if not newComplement.hasFeature(InternalFeature.DISCOURSE_FUNCTION):
                    newComplement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.OBJECT)
        elif isinstance(newComplement, str):
            newElement = StringElement(newComplement)
            complements.append(newElement)
            self.setFeature(InternalFeature.COMPLEMENTS, complements)
        else:
            raise ValueError('Invalid newComplement type: ' + str(type(newComplement)))

    # Sets a complement of the phrase element.
    def setComplement(self, newComplement):
        if isinstance(newComplement, NLGElement):
            function = newComplement.getFeature(InternalFeature.DISCOURSE_FUNCTION)
            self.removeComplements(function)
            self.addComplement(newComplement)
        elif isinstance(newComplement, str):
            self.setFeature(InternalFeature.COMPLEMENTS, None)
            self.addComplement(newComplement)
        else:
            raise ValueError('Invalid newComplement type: ' + str(type(newComplement)))

    # remove complements of the specified DiscourseFunction
    def removeComplements(self, function):
        complements = self.getFeatureAsElementList(InternalFeature.COMPLEMENTS)
        if function is None or complements is None:
            return
        complementsToRemove = []
        for complement in complements:
            if function == complement.getFeature(InternalFeature.DISCOURSE_FUNCTION):
                complementsToRemove.append(complement)
        if complementsToRemove:
            for compl in complementsToRemove:
                complements.remove(compl)
            self.setFeature(InternalFeature.COMPLEMENTS, complements)

    # Adds a new post-modifier to the phrase element.
    def addPostModifier(self, newPostModifier):
        postModifiers = self.getFeatureAsElementList(InternalFeature.POSTMODIFIERS)
        if postModifiers is None:
            postModifiers = []
        if isinstance(newPostModifier, NLGElement):
            newPostModifier.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.POST_MODIFIER)
            postModifiers.append(newPostModifier)
            self.setFeature(InternalFeature.POSTMODIFIERS, postModifiers)
        elif isinstance(newPostModifier, str):
            postModifiers.append(StringElement(newPostModifier))
            self.setFeature(InternalFeature.POSTMODIFIERS, postModifiers)
        else:
            raise ValueError('Invalid newPostModifier type: ' + str(type(newPostModifier)))

    # Set the postmodifier for this phrase.
    def setPostModifier(self, newPostModifier):
        if isinstance(newPostModifier, str) or isinstance(newPostModifier, NLGElement):
            self.setFeature(InternalFeature.POSTMODIFIERS, None)
            self.addPostModifier(newPostModifier)
        else:
            raise ValueError('Invalid newPostModifier type: ' + str(type(newPostModifier)))

    # Adds a new front modifier to the phrase element.
    def addFrontModifier(self, newFrontModifier):
        frontModifiers = self.getFeatureAsElementList(InternalFeature.FRONT_MODIFIERS)
        if frontModifiers is None:
            frontModifiers = []
        if isinstance(newFrontModifier, NLGElement):
            frontModifiers.append(newFrontModifier)
            self.setFeature(InternalFeature.FRONT_MODIFIERS, frontModifiers)
        elif isinstance(newFrontModifier, str):
            frontModifiers.append(StringElement(newFrontModifier))
            self.setFeature(InternalFeature.FRONT_MODIFIERS, frontModifiers)
        else:
            raise ValueError('Invalid newFrontModifier type: ' + str(type(newFrontModifier)))

    # Set the frontmodifier for this phrase. T
    def setFrontModifier(self, newFrontModifier):
        self.setFeature(InternalFeature.FRONT_MODIFIERS, None)
        self.addFrontModifier(newFrontModifier)

    # Adds a new pre-modifier to the phrase element.
    def addPreModifier(self, newPreModifier):
        if isinstance(newPreModifier, str):
            newPreModifier = StringElement(newPreModifier)
        preModifiers = self.getFeatureAsElementList(InternalFeature.PREMODIFIERS)
        if preModifiers is None:
            preModifiers = []
        preModifiers.append(newPreModifier)
        self.setFeature(InternalFeature.PREMODIFIERS, preModifiers)

    # Set the premodifier for this phrase.
    def setPreModifier(self, newPreModifier):
        self.setFeature(InternalFeature.PREMODIFIERS, None)
        self.addPreModifier(newPreModifier)

    # Add a modifier to a phrase Use heuristics to decide where it goes
    def addModifier(self, modifier):
        # default addModifier - always make modifier a preModifier
        if modifier is None:
            return
        # assume is preModifier, add in appropriate form
        self.addPreModifier(modifier)

    # Retrieves the current list of pre-modifiers for the phrase.
    def getPreModifiers(self):
        return self.getFeatureAsElementList(InternalFeature.PREMODIFIERS)

    # Retrieves the current list of post modifiers for the phrase.
    def getPostModifiers(self):
        return self.getFeatureAsElementList(InternalFeature.POSTMODIFIERS)

    # Retrieves the current list of frony modifiers for the phrase.
    def getFrontModifiers(self):
        return self.getFeatureAsElementList(InternalFeature.FRONT_MODIFIERS)

    # @Override
    def printTree(self, indent=None):
        if indent is None:
            indent = ''
        thisIndent      = indent + " |-"
        childIndent     = indent + " | "
        lastIndent      = indent + " \\-"
        lastChildIndent = indent + "   "
        pstr = "PhraseElement: category=" + str(self.getCategory()) + ", features={"
        features = self.getAllFeatures()
        for feat, val in features.items():
            pstr += feat + '=' + str(val) + ', '
        pstr = pstr[:-2]
        pstr += "}\n"
        children = self.getChildren()
        for i, child in enumerate(children):
            if i<len(children)-1:
                pstr += thisIndent + child.printTree(childIndent)
            else:
                pstr += lastIndent + child.printTree(lastChildIndent)
        return pstr

    # Removes all existing complements on the phrase.
    def clearComplements(self):
        self.removeFeature(InternalFeature.COMPLEMENTS)

    # Sets the determiner for the phrase.
    # @Deprecated
    def setDeterminer(self, newDeterminer):
        from .NLGFactory  import NLGFactory     # prevent circular imports
        factory = NLGFactory()
        determinerElement = factory.createWord(newDeterminer, LexicalCategory.DETERMINER)
        if determinerElement is not None:
            determinerElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.SPECIFIER)
            self.setFeature(InternalFeature.SPECIFIER, determinerElement)
            determinerElement.setParent(self)



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from .NLGElement                import *
from .LexicalCategory           import *
from ..features.Inflection      import *
from ..features.LexicalFeature  import *


# Internal class. This maintains inflectional variants of the word, which
# may be available in the lexicon.
class InflectionSet(object):
    def __init__(self, infl):
        self.infl  = infl   # Inflection
        self.forms = {}     # {str:str}  mapping values of LexicalFeature to actual word forms

    # set an inflectional form
    def addForm(self, feature, form):
        self.forms[feature] =  form

    # get an inflectional form
    def getForm(self, feature):
        return self.forms.get(feature, None)


# This is the class for a lexical entry (ie, a word).
class WordElement(NLGElement):
    # Words have baseForm, category, id, and features
    # features are inherited from NLGElement
    def __init__(self, baseForm=None, category=None, wid=None):
        super().__init__()
        if category is None:
            category = LexicalCategory.ANY
        self.setCategory(category)
        self.baseForm    = baseForm     # str
        self.id          = wid          # str (id in lexicon)
        self.inflVars    = {}           # {Inflection, InflectionSet} the inflectional variants
        self.defaultInfl = None         # Inflection # the default inflectional variant

    # Copy Constructor
    @staticmethod
    def fromWE(currentWord):
        we = WordElement(currentWord.getBaseForm(), currentWord.getCategory(), currentWord.getId())
        we.inflVars = currentWord.getInflectionalVariants()
        we.defaultInfl = currentWord.getDefaultInflectionalVariant()
        we.setFeatures(currentWord)
        return we

    #**********************************************************
    # getters and setters
    #**********************************************************
    def getBaseForm(self):
        return self.baseForm

    def getId(self):
        return self.id

    def setBaseForm(self, baseForm):
        self.baseForm = baseForm

    def setId(self, wid):
        self.id = wid

    # Set the default inflectional variant of a word.
    def setDefaultInflectionalVariant(self, variant):
        self.setFeature(LexicalFeature.DEFAULT_INFL, variant)
        self.defaultInfl = variant
        if variant in self.inflVars:
            infl_set = self.inflVars[variant]
            forms = LexicalFeature.getInflectionalFeatures(self.getCategory())
            if forms is not None:
                for f in forms:
                    self.setFeature(f, infl_set.getForm(f))

    # @return the default inflectional variant
    def getDefaultInflectionalVariant(self):
        return self.defaultInfl

    # Convenience method to get all the inflectional forms of the word.
    def getInflectionalVariants(self):
        return self.inflVars

    # Convenience method to set the default spelling variant of a word.
    def setDefaultSpellingVariant(self, variant):
        self.setFeature(LexicalFeature.DEFAULT_SPELL, variant)

    # Convenience method, equivalent to
    def getDefaultSpellingVariant(self):
        defSpell = self.getFeatureAsString(LexicalFeature.DEFAULT_SPELL)
        if defSpell is None:
            return self.getBaseForm()
        return defSpell

    # Add an inflectional variant to this word element.
    def addInflectionalVariant(self, infl, lexicalFeature=None, form=None):
        if lexicalFeature is None and form is None:
            self.inflVars[infl] = InflectionSet(infl)
        elif infl in self.inflVars:
            self.inflVars[infl].addForm(lexicalFeature, form)
        else:
            infl_set = InflectionSet(infl)
            infl_set.addForm(lexicalFeature, form)
            self.inflVars[infl] = infl_set

    # Check whether this word has a particular inflectional variant
    def hasInflectionalVariant(self, infl):
        return infl in self.inflVars

    # Sets Features from another existing WordElement into this WordElement.
    def setFeatures(self, currentWord):
        if currentWord is not None and currentWord.getAllfeatures():
            for feature in currentWord.getAllFeatureNames():
                self.setFeature(feature, currentWord.getFeature(feature))

    #**********************************************************
    # other methods
    #**********************************************************
    def __str__(self):
        buffer = 'WordElement[' + self.getBaseForm() + ':'
        category = self.getCategory()
        if category:
            buffer += str(category)
        else:
            buffer += "no category"
        buffer += ']'
        return buffer

    # This method returns an empty List as words do not have child elements.
    def getChildren(self):
        return []

    def printTree(self, indent=None):
        pstr = "WordElement: base=" + self.getBaseForm() + ", category=" + str(self.getCategory()) + ", " \
             + super().__str__() + '\n'
        return pstr

    # Check if this WordElement is equal to an object.
    def __eq__(self, we):
        if isinstance(we, WordElement):
            return self.baseForm==we.baseForm and self.id==we.id and self.features==we.features
        return False
    def __ne__(self, we):
        return not self.__eq__(we)

    # Defining __eq__ removes the default hash
    def __hash__(self):
        items = [self.baseForm, self.id]
        for key, value in self.features.items():
            items.append( (key, value) )
        return hash( tuple(items) )



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from .DocumentCategory  import *
from .NLGElement        import *

# DocumentElement is a convenient extension of the base NLGElement class.
class DocumentElement(NLGElement):
    FEATURE_TITLE       = "textTitle"
    FEATURE_COMPONENTS  = "textComponents"

    def __init__(self, category=None, textTitle=None):
        super().__init__()
        if category is not None:
            self.setCategory(category)
        if textTitle is not None:
            self.setTitle(textTitle)

    # Sets the title of this element. Titles are specifically used with
    # documents (the document name) and sections (headings).
    def setTitle(self, textTitle):
        self.setFeature(self.FEATURE_TITLE, textTitle)

    # Retrieves the title of this element.
    def getTitle(self):
        return self.getFeatureAsString(self.FEATURE_TITLE)

    # Retrieves the child components of this element.
    def getComponents(self):
        return self.getFeatureAsElementList(self.FEATURE_COMPONENTS)

    # Add a single child component to the current list of child components. If
    # there are no existing child components a new list is created.
    def addComponent(self, element):
        if element is None:
            return
        thisCategory = self.getCategory()
        category = element.getCategory()
        if category is not None and isinstance(thisCategory, DocumentCategory):
            if thisCategory.hasSubPart(category):
                self.addElementToComponents(element)
            else:
                promotedElement = self.promote(element)
                if promotedElement is not None:
                    self.addElementToComponents(promotedElement)
                else: # error condition - add original element so something is visible
                    self.addElementToComponents(element)
        else:
            self.addElementToComponents(element)

    # add an element to a components list
    def addElementToComponents(self, element):
        components = self.getComponents()
        components.append(element)
        element.setParent(self)
        self.setComponents(components)

    # promote an NLGElement so that it is at the right level to be added to a DocumentElement/
    # Promotion means adding surrounding nodes at higher doc levels
    def promote(self, element):
        # check if promotion needed
        if self.getCategory().hasSubPart(element.getCategory()):
            return element
        # if element is not a DocumentElement, embed it in a sentence and recurse
        if not isinstance(element, DocumentElement):
            sentence = DocumentElement(DocumentCategory.SENTENCE, None)
            sentence.addElementToComponents(element)
            return self.promote(sentence)
        # if element is a Sentence, promote it to a paragraph
        if element.getCategory() == DocumentCategory.SENTENCE:
            paragraph = DocumentElement(DocumentCategory.PARAGRAPH, None)
            paragraph.addElementToComponents(element)
            return self.promote(paragraph)
        return None

    # Adds a collection of NLGElements to the list of child
    # components. If there are no existing child components, then a new list is
    # created.
    def addComponents(self, textComponents):
        if textComponents is None:
            return
        thisCategory = self.getCategory()
        elementsToAdd = []
        category = None
        for eachElement in textComponents:
            if isinstance(eachElement, NLGElement):
                category = eachElement.getCategory()
                if category is not None and isinstance(thisCategory, DocumentCategory):
                    if thisCategory.hasSubPart(category):
                        elementsToAdd.append(eachElement)
                        eachElement.setParent(self)
        if len(elementsToAdd)> 0:
            components = self.getComponents()
            if components is None:
                components = []
            components.extend(elementsToAdd)
            self.setFeature(self.FEATURE_COMPONENTS, components)

    # Removes the specified component from the list of child components.
    def removeComponent(self, textComponent):
        components = self.getComponents()
        if components:
            if textComponent in components:
                components.remove(textComponent)
                return True
        return False

    # Removes all the child components from this element.
    def clearComponents(self):
        components = self.getComponents()
        if components is not None:
            components.clear()

    # Child elements of a DocumentElement are the components.
    # @Override
    def getChildren(self):
        return self.getComponents()

    # Replaces the existing components with the supplied list of components.
    # This is identical to calling:
    def setComponents(self, components):
        self.setFeature(self.FEATURE_COMPONENTS, components)

    # @Override
    def printTree(self, indent=None):
        if indent is None:
            indent = ''
        thisIndent      = indent + " |-"
        childIndent     = indent + " | "
        lastIndent      = indent + " \\-"
        lastChildIndent = indent + "   "
        pstr = "DocumentElement: category=" + str(self.getCategory())
        realisation = self.getRealisation()
        if realisation:
            pstr += " realisation=" + realisation
        pstr += '\n'
        children = self.getChildren()
        for i, child in enumerate(children):
            if i < len(children)-1:
                pstr += thisIndent + child.printTree(childIndent)
            else:
                pstr += lastIndent + child.printTree(lastChildIndent)
        return pstr



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ...features.DiscourseFunction      import *
from ...features.Feature                import *
from ...features.InternalFeature        import *
from ...features.LexicalFeature         import *
from ...framework.InflectedWordElement  import *
from ...framework.LexicalCategory       import *
from ...framework.ListElement           import *
from ...framework.PhraseCategory        import *
from ...framework.PhraseElement         import *


# This class contains static methods to help the syntax processor realise
# phrases.
class PhraseHelper(object):
    # The main method for realising phrases.
    @classmethod
    def realise(cls, parent, phrase):
        realisedElement = ListElement()
        if phrase is not None:
            realisedElement = ListElement()
            cls.realiseList(parent, realisedElement, phrase.getPreModifiers(), \
                    DiscourseFunction.PRE_MODIFIER)
            cls.realiseHead(parent, phrase, realisedElement)
            cls.realiseComplements(parent, phrase, realisedElement)
            PhraseHelper.realiseList(parent, realisedElement, \
                    phrase.getPostModifiers(), DiscourseFunction.POST_MODIFIER)
        return realisedElement

    # Realises the complements of the phrase adding and where
    # appropriate.
    @classmethod
    def realiseComplements(cls, parent, phrase, realisedElement):
        firstProcessed = False
        for complement in phrase.getFeatureAsElementList(InternalFeature.COMPLEMENTS):
            currentElement = parent.realise(complement)
            if currentElement is not None:
                currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.COMPLEMENT)
                if firstProcessed:
                    realisedElement.addComponent(InflectedWordElement("and", LexicalCategory.CONJUNCTION))
                else:
                    firstProcessed = True
                realisedElement.addComponent(currentElement)

    # Realises the head element of the phrase.
    @classmethod
    def realiseHead(cls, parent, phrase, realisedElement):
        head = phrase.getHead()
        if head is not None:
            if phrase.hasFeature(Feature.IS_COMPARATIVE):
                head.setFeature(Feature.IS_COMPARATIVE, phrase.getFeature(Feature.IS_COMPARATIVE))
            elif phrase.hasFeature(Feature.IS_SUPERLATIVE):
                head.setFeature(Feature.IS_SUPERLATIVE, phrase.getFeature(Feature.IS_SUPERLATIVE))
            head = parent.realise(head)
            head.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.HEAD)
            realisedElement.addComponent(head)

    # Iterates through a List of NLGElements
    # realisation each element and adding it to the on-going realisation of
    # this clause.
    @classmethod
    def realiseList(cls, parent, realisedElement, elementList, function):
        # AG: Change here: the original list structure is kept, i.e. rather
        # than taking the elements of the list and putting them in the realised
        # element, we now add the realised elements to a new list and put that
        # in the realised element list. This preserves constituency for
        # orthography and morphology processing later.
        realisedList = ListElement()
        for eachElement in elementList:
            currentElement = parent.realise(eachElement)
            if currentElement is not None:
                currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, function)
                if eachElement.getFeatureAsBoolean(Feature.APPOSITIVE):
                    currentElement.setFeature(Feature.APPOSITIVE, True)
                realisedList.addComponent(currentElement)
        if realisedList.getChildren():
            realisedElement.addComponent(realisedList)

    # Determines if the given phrase has an expletive as a subject.
    @classmethod
    def isExpletiveSubject(cls, phrase):
        subjects = phrase.getFeatureAsElementList(InternalFeature.SUBJECTS)
        expletive = False
        if len(subjects) == 1:
            subjectNP = subjects[0]
            if subjectNP.isA(PhraseCategory.NOUN_PHRASE):
                expletive = subjectNP.getFeatureAsBoolean(LexicalFeature.EXPLETIVE_SUBJECT)
            elif subjectNP.isA(PhraseCategory.CANNED_TEXT):
                expletive = "there" == subjectNP.getRealisation().lower()
        return expletive



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ..features.DiscourseFunction           import *
from ..features.Feature                     import *
from ..features.Form                        import *
from ..features.InternalFeature             import *
from ..features.Person                      import *
from ..features.Tense                       import *
from ..framework.CoordinatedPhraseElement   import *
from ..framework.InflectedWordElement       import *
from ..framework.LexicalCategory            import *
from ..framework.NLGElement                 import *
from ..framework.PhraseCategory             import *
from ..framework.PhraseElement              import *
from ..framework.WordElement                import *


# This class defines a verb phrase.
class VPPhraseSpec(PhraseElement):
    def __init__(self, phraseFactory):
        super().__init__(PhraseCategory.VERB_PHRASE)
        self.setFactory(phraseFactory)
        # set default feature values
        self.setFeature(Feature.PERFECT, False)
        self.setFeature(Feature.PROGRESSIVE, False)
        self.setFeature(Feature.PASSIVE, False)
        self.setFeature(Feature.NEGATED, False)
        self.setFeature(Feature.TENSE, Tense.PRESENT)
        self.setFeature(Feature.PERSON, Person.THIRD)
        self.setPlural(False)
        self.setFeature(Feature.FORM, Form.NORMAL)
        self.setFeature(InternalFeature.REALISE_AUXILIARY, True)

    # sets the verb (head) of a verb phrase.
    def setVerb(self, verb):
        if isinstance(verb, str):
            try:
                space = verb.index(' ')
            except ValueError:
                space = -1
            if space < 0: # no space, so no particle
                verbElement = self.getFactory().createWord(verb, LexicalCategory.VERB)
            else: # space, so break up into verb and particle
                verbElement = self.getFactory().createWord(verb[:space], LexicalCategory.VERB)
                self.setFeature(Feature.PARTICLE, verb[space+1:])
        else:  # Object is not a String
            verbElement = self.getFactory().createNLGElement(verb,LexicalCategory.VERB)
        self.setHead(verbElement)

    # @return verb (head) of verb phrase
    def getVerb(self):
        return self.getHead()

    # Sets the direct object of a clause  (assumes this is the only direct object)
    def setObject(self, obj):
        if isinstance(obj, PhraseElement) or isinstance(obj, CoordinatedPhraseElement):
            objectPhrase = obj
        else:
            objectPhrase = self.getFactory().createNounPhrase(obj)
        objectPhrase.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.OBJECT)
        self.setComplement(objectPhrase)

    # Returns the direct object of a clause (assumes there is only one)
    def getObject(self):
        complements = self.getFeatureAsElementList(InternalFeature.COMPLEMENTS)
        for complement in complements:
            if complement.getFeature(InternalFeature.DISCOURSE_FUNCTION) == DiscourseFunction.OBJECT:
                return complement
        return None

    # Set the indirect object of a clause (assumes this is the only direct indirect object)
    def setIndirectObject(self, indirectObject):
        if isinstance(indirectObject, PhraseElement) or isinstance(indirectObject, CoordinatedPhraseElement):
            indirectObjectPhrase = indirectObject
        else:
            indirectObjectPhrase = self.getFactory().createNounPhrase(indirectObject)
        indirectObjectPhrase.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.INDIRECT_OBJECT)
        self.setComplement(indirectObjectPhrase)

    # Returns the indirect object of a clause (assumes there is only one)
    def getIndirectObject(self):
        complements = self.getFeatureAsElementList(InternalFeature.COMPLEMENTS)
        for complement in complements:
            if complement.getFeature(InternalFeature.DISCOURSE_FUNCTION) == DiscourseFunction.INDIRECT_OBJECT:
                return complement
        return None

    # Add a modifier to a verb phrase
    # @Override
    def addModifier(self, modifier):
        # adverb is preModifier
        # string which is one lexicographic word is looked up in lexicon,
        # if it is an adverb than it becomes a preModifier
        # Everything else is postModifier
        if modifier is None:
            return
        # get modifier as NLGElement if possible
        if isinstance(modifier, NLGElement):
            modifierElement = modifier
        elif isinstance(modifier, str):
            modifierString = modifier
            if len(modifierString)>0 and " " not in modifierString:
                modifierElement = self.getFactory().createWord(modifier, LexicalCategory.ANY)
        # if no modifier element, must be a complex string
        if modifierElement is None:
            self.addPostModifier(modifier)
            return
        # extract WordElement if modifier is a single word
        if isinstance(modifierElement, WordElement):
            modifierWord = modifierElement
        elif isinstance(modifierElement, InflectedWordElement):
            modifierWord = modifierElement.getBaseWord()
        if modifierWord is not None and modifierWord.getCategory() == LexicalCategory.ADVERB:
            self.addPreModifier(modifierWord)
            return
        # default case
        self.addPostModifier(modifierElement)



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from abc import ABC
from ..lexicon.Lexicon  import *


# NLGModule is the base class that all processing modules extend
# from.
class NLGModule(ABC):
    def __init__(self):
        self.lexicon = None    #Lexicon

    # Performs one-time initialisation of the module.
    def initialise(self):
        assert False, 'Not implemented in base.'

    # Realises the given element.
    def realise(self, element):
        assert False, 'Not implemented in base.'

    # Sets the lexicon to be used by this module. Passing in null
    # will remove the existing lexicon and no lexicon will be used.
    def setLexicon(self, newLexicon):
        self.lexicon = newLexicon

    # Retrieves the lexicon currently being used by this module.
    def getLexicon(self):
        return self.lexicon



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from abc            import ABC
from ..framework.LexicalCategory    import *
from ..framework.WordElement        import *


# This is the generic abstract class for a Lexicon.
class Lexicon(ABC):
    def __init__(self):
        pass

    # Return an instance of the default lexicon
    @staticmethod
    def getDefaultLexicon():
        from .XMLLexicon import XMLLexicon  #import here to prevent circular import
        return XMLLexicon()

    # create a default WordElement. May be overridden by specific types of lexicon
    def createWord(self, baseForm, category=None):
        return WordElement(baseForm, category)

    # General word lookup method, tries base form, variant, ID (in this order)
    # Creates new word if can't find existing word
    def lookupWord(self, baseForm, category=None):
        if not category:
            category = LexicalCategory.ANY
        if self.hasWord(baseForm, category):
            return self.getWord(baseForm, category)
        elif self.hasWordFromVariant(baseForm, category):
            return self.getWordFromVariant(baseForm, category)
        elif self.hasWordByID(baseForm):
            return self.getWordByID(baseForm)
        else:
            return self.createWord(baseForm, category)


    # returns all Words which have the specified base form and category
    def getWords(self, baseForm, category):
        assert False, 'Not implemented in base.'

    # get a WordElement which has the specified base form and category
    def getWord(self, baseForm, category=None):
        wordElements = self.getWords(baseForm, category)
        if not wordElements:
            return self.createWord(baseForm, category)
        else:
            return self.selectMatchingWord(wordElements, baseForm)

    # choose a single WordElement from a list of WordElements.  Prefer one
    # which exactly matches the baseForm
    def selectMatchingWord(self, wordElements, baseForm):
        # EHUD REITER  - this method added because some DBs are case-insensitive,
        # so a query on "man" returns both "man" and "MAN".  In such cases, the
        # exact match (eg, "man") should be returned
        # below check is redundant, since caller should check this
        if not wordElements:
            return self.createWord(baseForm)
        # look for exact match in base form
        for wordElement in wordElements:
            if wordElement.getBaseForm()==baseForm:
                return wordElement
        # Roman Kutlak: I don't think it is a good idea to return a word whose
        # case does not match because if a word appears in the lexicon
        # as an acronym only, it will be replaced as such. For example,
        # "foo" will return as the acronym "FOO". This does not seem desirable.
        # else return first element in list
        if wordElements[0].getBaseForm().lower() == baseForm.lower():
            return self.createWord(baseForm, LexicalCategory.ANY)
        return wordElements[0]

    # return true if the lexicon contains a WordElement which has
    # the specified base form and category
    def hasWord(self, baseForm, category=None):
        return len(self.getWords(baseForm, category))>0

    # returns a List of WordElement which have this ID.
    def getWordsByID(self, wid):
        assert False, 'Not implemented in base.'

    # get a WordElement with the specified ID
    def getWordByID(self, wid):
        wordElements = self.getWordsByID(wid)
        if not wordElements:
            return self.createWord(wid)
        else:
            return wordElements[0]

    # return true if the lexicon contains a WordElement which the specified ID
    def hasWordByID(self, wid):
        return len(self.getWordsByID(wid))>0


    # returns Words which have an inflected form and/or spelling variant that
    # matches the specified variant, and are in the specified category.
    def getWordsFromVariant(self, variant, category):
        assert False, 'Not implemented in base.'

    # returns a WordElement which has the specified inflected form and/or
    # spelling variant that matches the specified variant, of the specified
    # category
    def getWordFromVariant(self, variant, category=None):
        wordElements = self.getWordsFromVariant(variant, category)
        if not wordElements:
            return self.createWord(variant, category)
        if category is not None:
            return self.selectMatchingWord(wordElements, variant)
        else:
            return wordElements[0]

    # return true if the lexicon contains a WordElement which
    # matches the specified variant form and category
    def hasWordFromVariant(self, variant, category=None):
        return len(self.getWordsFromVariant(variant, category))>0

    #close the lexicon (if necessary) if lexicon does not need to be closed, this does nothing
    def close(self):
        pass



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from .NLGElement                import *
from ..features.InternalFeature import *


# ListElement is used to define elements that can be grouped
# together and treated in a similar manner.
class ListElement(NLGElement):
    def __init__(self, component=None):
        super().__init__()
        if component is None:
            return
        elif isinstance(component, list):
            self.addComponents(component)
        elif isinstance(component, NLGElement):
            self.addComponent(component)
        else:
            raise ValueError('Invalid component type: ' + str(type(component)))

    # @Override
    def getChildren(self):
        return self.getFeatureAsElementList(InternalFeature.COMPONENTS)

    # Adds the given component to the list element.
    def addComponent(self, newComponent):
        components = self.getFeatureAsElementList(InternalFeature.COMPONENTS)
        if components is None:
            components = []
        self.setFeature(InternalFeature.COMPONENTS, components)
        components.append(newComponent)

    # Adds the given components to the list element.
    def addComponents(self, newComponents):
        components = self.getFeatureAsElementList(InternalFeature.COMPONENTS)
        if components is None:
            components = []
        self.setFeature(InternalFeature.COMPONENTS, components)
        components.extend(newComponents)

    # Replaces the current components in the list element with the given list.
    def setComponents(self, newComponents):
        self.setFeature(InternalFeature.COMPONENTS, newComponents)

    # @Override
    def __str__(self):
        pstr = '['
        for child in self.getChildren():
            pstr += str(child) + ', '
        pstr = pstr[:-2] + ']'
        return pstr

    # @Override
    def printTree(self, indent=None):
        if indent is None:
            indent = ''
        thisIndent      = indent + " |-"
        childIndent     = indent + " | "
        lastIndent      =  indent + " \\-"
        lastChildIndent = indent + "   "
        pstr = "ListElement: features={"
        features = self.getAllFeatures()
        for feat, val in features.items():
            pstr += feat + '=' + str(val) + ' '
        pstr += "}\n"
        children = self.getChildren()
        for i, child in enumerate(children):
            if i<len(children)-1:
                pstr += thisIndent + child.printTree(childIndent)
            else:
                pstr += lastIndent + child.printTree(lastChildIndent)
        return pstr

    # Retrieves the number of components in this list element.
    # @return the number of components.
    def __len__(self):
        return len(self.getChildren())

    # Retrieves the first component in the list.
    # @return the NLGElement at the top of the list.
    def getFirst(self):
        children = self.getChildren()
        if not children:
            return None
        return children[0]



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ...format.english.TextFormatter                import *
from ...framework.DocumentCategory                  import *
from ...framework.DocumentElement                   import *
from ...framework.NLGElement                        import *
from ...framework.NLGModule                         import *
from ...morphology.english.MorphologyProcessor      import *
from ...orthography.english.OrthographyProcessor    import *
from ...syntax.english.SyntaxProcessor              import *


class Realiser(NLGModule):
    def __init__(self, lexicon=None):
        super().__init__()
        self.initialise()
        if lexicon is not None:
            self.setLexicon(lexicon)
        self.debug = False

    # Check whether this processor separates premodifiers using a comma.
    def isCommaSepPremodifiers(self):
        if self.orthography is None:
            return False
        return self.orthography.isCommaSepPremodifiers()

    # Set whether to separate premodifiers using a comma.
    def setCommaSepPremodifiers(self, commaSepPremodifiers):
        if self.orthography is not None:
            self.orthography.setCommaSepPremodifiers(commaSepPremodifiers)

    # Check whether this processor separates cue phrases from the matrix clause using a comma.
    def isCommaSepCuephrase(self):
        if self.orthography is None:
            return False
        return self.orthography.isCommaSepCuephrase()

    # Set whether to separate cue phrases from the host phrase using a comma.
    def setCommaSepCuephrase(self, commaSepCuephrase):
        if self.orthography is not None:
            self.orthography.setCommaSepCuephrase(commaSepCuephrase)

    # @Override
    def initialise(self):
        self.morphology = MorphologyProcessor()
        self.morphology.initialise()
        self.orthography = OrthographyProcessor()
        self.orthography.initialise()
        self.syntax = SyntaxProcessor()
        self.syntax.initialise()
        self.formatter = TextFormatter()
        self.formatter.initialise()

    # @Override
    def realise(self, element):
        if isinstance(element, NLGElement):
            return self._realiseElement(element)
        elif isinstance(element, list):
            return self._realiseElementList(element)
        elif element is None:
            return ListElement()
        else:
            raise ValueError('Invalid element type: ' + str(type(element)))

    def _realiseElement(self, element):
        if self.debug:
            print("INITIAL TREE")
            print(element.printTree(None) + '\n')
        postSyntax = self.syntax.realise(element)
        if self.debug:
            print("POST-SYNTAX TREE")
            print(postSyntax.printTree(None) + '\n')
        postMorphology = self.morphology.realise(postSyntax)
        if self.debug:
            print("POST-MORPHOLOGY TREE")
            print(postMorphology.printTree(None) + '\n')
        postOrthography = self.orthography.realise(postMorphology)
        if self.debug:
            print("POST-ORTHOGRAPHY TREE")
            print(postOrthography.printTree(None) + '\n')
        if self.formatter is not None:
            postFormatter = self.formatter.realise(postOrthography)
            if self.debug:
                print("POST-FORMATTER TREE")
                print(postFormatter.printTree(None) + '\n')
        else:
            postFormatter = postOrthography
        return postFormatter

    # Convenience class to realise any NLGElement as a sentence
    def realiseSentence(self, element):
        if isinstance(element, DocumentElement):
            realised = self.realise(element)
        else:
            sentence = DocumentElement(DocumentCategory.SENTENCE, None)
            sentence.addComponent(element)
            realised = self.realise(sentence)
        if realised is None:
            return None
        else:
            return realised.getRealisation()

    # @Override
    def _realiseElementList(self, elements):
        realisedElements = []
        if elements is not None:
            for element in elements:
                realisedElement = self.realise(element)
                realisedElements.append(realisedElement)
        return realisedElements

    # @Override
    def setLexicon(self, newLexicon):
        self.syntax.setLexicon(newLexicon)
        self.morphology.setLexicon(newLexicon)
        self.orthography.setLexicon(newLexicon)

    def setFormatter(self, formatter):
        self.formatter = formatter

    def setDebugMode(self, debugOn):
        self.debug = debugOn



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ..features.DiscourseFunction   import *
from ..features.InternalFeature     import *
from ..framework.LexicalCategory    import *
from ..framework.NLGElement         import *
from ..framework.PhraseCategory     import *
from ..framework.PhraseElement      import *


# This class defines a prepositional phrase.  
class PPPhraseSpec(PhraseElement):
    def __init__(self, phraseFactory):
        super().__init__(PhraseCategory.PREPOSITIONAL_PHRASE)
        self.setFactory(phraseFactory)

    # sets the preposition (head) of a prepositional phrase
    def setPreposition(self, preposition):
        if isinstance(preposition, NLGElement):
            self.setHead(preposition)
        else:
            # create noun as word
            prepositionalElement = self.getFactory().createWord(preposition, LexicalCategory.PREPOSITION)
            # set head of NP to nounElement
            self.setHead(prepositionalElement)

    # @return preposition (head) of prepositional phrase
    def getPreposition(self):
        return self.getHead()

    # Sets the  object of a PP
    def setObject(self, obj):
        objectPhrase = self.getFactory().createNounPhrase(obj)
        objectPhrase.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.OBJECT)
        self.addComplement(objectPhrase)

    # @return object of PP (assume only one)
    def getObject(self):
        complements = self.getFeatureAsElementList(InternalFeature.COMPLEMENTS)
        for complement in complements:
            if complement.getFeature(InternalFeature.DISCOURSE_FUNCTION) == DiscourseFunction.OBJECT:
                return complement
        return None



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from enum import Enum

# This enumeration defines the different syntactical phrases.
class PhraseCategory(Enum):
    # A grammatical clause, the simplest form of which consists of a subject
    # (noun or noun phrase) and a verb (or verb phrase).
    CLAUSE = 0
    # A phrase relating to an adjective.
    ADJECTIVE_PHRASE = 1
    # A phrase relating to an adverb.
    ADVERB_PHRASE = 2
    # A phrase relating to a noun.
    NOUN_PHRASE = 3
    # A phrase relating to a preposition.
    PREPOSITIONAL_PHRASE = 4
    # A phrase relating to a verb.
    VERB_PHRASE = 5
    # A phrase relating to a pre-formed string that is not altered in anyway.#/
    CANNED_TEXT = 6

    # Checks to see if the given object is equal to this document category
    def equalTo(self, checkObject):
        match = False
        if checkObject is not None:
            if isinstance(checkObject, PhraseCategory):
                match = self==checkObject
            else:
                match = str(self.name).lower() == str(checkObject).lower()
        return match

    def __str__(self):
        return self.name



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from enum import Enum

# An enumeration representing the different types of number agreement.
class NumberAgreement(Enum):
    # This represents words that have the same form regardless of whether they
    # are singular or plural.
    BOTH = 0
    # This represents verbs and nouns that are written in the plural.
    PLURAL = 1
    # This represents verbs and nouns that are written in the singular. Fo
    SINGULAR  = 2

    def __str__(self):
        return self.name



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

import re
from .CoordinatedPhraseElement  import *
from .DocumentCategory          import *
from .DocumentElement           import *
from .InflectedWordElement      import *
from .NLGElement                import *
from .PhraseElement             import *
from .StringElement             import *
from .WordElement               import *
from ..features.Feature         import *
from ..features.Gender          import *
from ..features.InternalFeature import *
from ..features.LexicalFeature  import *
from ..features.NumberAgreement import *
from ..features.Person          import *
from ..lexicon.Lexicon          import *
from ..phrasespec               import *


# This class contains methods for creating syntactic phrases. These methods
# should be used instead of directly invoking new on SPhraseSpec, etc.
class NLGFactory(object):
    PRONOUNS = ["I", "you", "he", "she", "it", "me", "you", "him", "her", "it", \
                "myself", "yourself", "himself", "herself", "itself", "mine", \
                "yours", "his", "hers", "its", "we", "you", "they", "they", "they", \
                "us", "you", "them", "them", "them", "ourselves", "yourselves", \
                "themselves", "themselves", "themselves", "ours", "yours", "theirs", \
                "theirs", "theirs", "there"]
    FIRST_PRONOUNS = ["I", "me", "myself", "we", "us", "ourselves", "mine", "my", "ours", "our"]
    SECOND_PRONOUNS = ["you","yourself", "yourselves", "yours", "your"]
    REFLEXIVE_PRONOUNS = ["myself", "yourself", "himself", "herself", "itself", \
                          "ourselves", "yourselves", "themselves"]
    MASCULINE_PRONOUNS = ["he", "him", "himself", "his"]
    FEMININE_PRONOUNS = ["she", "her", "herself", "hers"]
    POSSESSIVE_PRONOUNS =["mine", "ours", "yours", "his", "hers", "its", "theirs", \
                          "my", "our", "your", "her", "their"]
    PLURAL_PRONOUNS = ["we", "us", "ourselves", "ours", "our", "they", "them", "theirs", "their"]
    EITHER_NUMBER_PRONOUNS = ["there"]
    EXPLETIVE_PRONOUNS = ["there"]
    WORD_REGEX = r"\w*" # regex for determining if a string is a single word or not

    # Creates a new phrase factory with the associated lexicon.
    def __init__(self, newLexicon=None):
        self.setLexicon(newLexicon)

    # Sets the lexicon to be used by this factory. Passing a parameter of
    # null will remove any existing lexicon from the factory.
    def setLexicon(self, newLexicon):
        self.lexicon = newLexicon

    # Creates a new element representing a word. If the word passed is already
    # an NLGElement then that is returned unchanged.
    def createWord(self, word, category):
        wordElement = None
        if isinstance(word, NLGElement):
            wordElement = word
        elif isinstance(word, str) and self.lexicon is not None:
            wordElement = self.lexicon.lookupWord(word, category)
            if word in self.PRONOUNS:
                self.setPronounFeatures(wordElement, word)
        return wordElement

    # Create an inflected word element.
    def createInflectedWord(self, word, category):
        inflElement = None
        if isinstance(word, WordElement):
            inflElement = InflectedWordElement(word)
        elif isinstance(word, str):
            baseword = self.createWord(word, category)
            if baseword is not None and isinstance(baseword, WordElement):
                inflElement = InflectedWordElement(baseword)
            else:
                inflElement = InflectedWordElement(word, category)
        elif isinstance(word, NLGElement):
            inflElement = word
        return inflElement

    # A helper method to set the features on newly created pronoun words.
    def setPronounFeatures(self, wordElement, word):
        wordElement.setCategory(LexicalCategory.PRONOUN)
        if word in self.FIRST_PRONOUNS:
            wordElement.setFeature(Feature.PERSON, Person.FIRST)
        elif word in self.SECOND_PRONOUNS:
            wordElement.setFeature(Feature.PERSON, Person.SECOND)
            if "yourself" == word.lower():
                wordElement.setPlural(False)
            elif "yourselves" == word.lower():
                wordElement.setPlural(True)
            else:
                wordElement.setFeature(Feature.NUMBER, NumberAgreement.BOTH)
        else:
            wordElement.setFeature(Feature.PERSON, Person.THIRD)
        if word in self.REFLEXIVE_PRONOUNS:
            wordElement.setFeature(LexicalFeature.REFLEXIVE, True)
        else:
            wordElement.setFeature(LexicalFeature.REFLEXIVE, False)
        if word in self.MASCULINE_PRONOUNS:
            wordElement.setFeature(LexicalFeature.GENDER, Gender.MASCULINE)
        elif word in self.FEMININE_PRONOUNS:
            wordElement.setFeature(LexicalFeature.GENDER, Gender.FEMININE)
        else:
            wordElement.setFeature(LexicalFeature.GENDER, Gender.NEUTER)
        if word in self.POSSESSIVE_PRONOUNS:
            wordElement.setFeature(Feature.POSSESSIVE, True)
        else:
            wordElement.setFeature(Feature.POSSESSIVE, False)
        if word in self.PLURAL_PRONOUNS and not word in self.SECOND_PRONOUNS:
            wordElement.setPlural(True)
        elif word not in self.EITHER_NUMBER_PRONOUNS:
            wordElement.setPlural(False)
        if word in self.EXPLETIVE_PRONOUNS:
            wordElement.setFeature(InternalFeature.NON_MORPH, True)
            wordElement.setFeature(LexicalFeature.EXPLETIVE_SUBJECT, True)

    # Creates a preposition phrase with the given preposition and complement.
    def createPrepositionPhrase(self, preposition=None, complement=None):
        phraseElement = PPPhraseSpec(self)
        prepositionalElement = self.createNLGElement(preposition, LexicalCategory.PREPOSITION)
        self.setPhraseHead(phraseElement, prepositionalElement)
        if complement is not None:
            self.setComplement(phraseElement, complement)
        return phraseElement

    # A helper method for setting the complement of a phrase.
    def setComplement(self, phraseElement, complement):
        complementElement = self.createNLGElement(complement)
        phraseElement.addComplement(complementElement)

    # this method creates an NLGElement from an object
    def createNLGElement(self, element, category=None):
        if element is None:
            return None
        if category is None:
            category = LexicalCategory.ANY
        # InflectedWordElement - return underlying word
        if isinstance(element, InflectedWordElement):
            return element.getBaseWord()
        # StringElement - look up in lexicon if it is a word otherwise return element
        elif isinstance(element, StringElement):
            if self.stringIsWord(element.getRealisation(), category):
                return self.createWord(element.getRealisation(), category)
            else:
                return element
        # other NLGElement - return element
        elif isinstance(element, NLGElement):
            return element
        # String - look up in lexicon if a word, otherwise return StringElement
        elif isinstance(element, str):
            if self.stringIsWord(element, category):
                return self.createWord(element, category)
            else:
                return StringElement(element)
        else:
            raise ValueError('Invalid element type: ' + str(type(element)))

    # return true if string is a word
    def stringIsWord(self, string, category):
        return self.lexicon is not None and \
               (self.lexicon.hasWord(string, category) or string in self.PRONOUNS or \
                re.fullmatch(self.WORD_REGEX, string))

    # Creates a noun phrase with the given specifier and subject.
    def createNounPhrase(self, ina=None, inb=None):
        # single paramter => createNounPhrase(noun)
        if ina is not None and inb is None:
            return self._createNP1(ina)
        # otherwise assume two paramters => createNounPhrase(specifier, noun)
        else:
            return self._createNP2(ina, inb)

    # createNounPhrase(noun)
    def _createNP1(self, noun):
        if isinstance(noun, NPPhraseSpec):
            return noun
        else:
            return self._createNP2(None, noun)

    # createNounPhrase(sepcifer, noun)
    def _createNP2(self, specifier, noun):
        if isinstance(noun, NPPhraseSpec):
            return noun
        phraseElement = NPPhraseSpec(self)
        nounElement = self.createNLGElement(noun, LexicalCategory.NOUN)
        self.setPhraseHead(phraseElement, nounElement)
        if specifier is not None:
            phraseElement.setSpecifier(specifier)
        return phraseElement

    # A helper method to set the head feature of the phrase.
    def setPhraseHead(self, phraseElement , headElement):
        if headElement is not None:
            phraseElement.setHead(headElement)
            headElement.setParent(phraseElement)

    # Creates an adjective phrase wrapping the given adjective.
    def createAdjectivePhrase(self, adjective=None):
        phraseElement = AdjPhraseSpec(self)
        adjectiveElement = self.createNLGElement(adjective, LexicalCategory.ADJECTIVE)
        self.setPhraseHead(phraseElement, adjectiveElement)
        return phraseElement

    # Creates a verb phrase wrapping the main verb given.
    def createVerbPhrase(self, verb=None):
        phraseElement = VPPhraseSpec(self)
        phraseElement.setVerb(verb)
        self.setPhraseHead(phraseElement, phraseElement.getVerb())
        return phraseElement

    # Creates an adverb phrase wrapping the given adverb.
    def createAdverbPhrase(self, adverb=None):
        phraseElement = AdvPhraseSpec(self)
        adverbElement = self.createNLGElement(adverb, LexicalCategory.ADVERB)
        self.setPhraseHead(phraseElement, adverbElement)
        return phraseElement

    # Creates a clause with the given subject, verb or verb phrase and direct
    # object but no indirect object.
    def createClause(self, subject=None, verb=None, directObject=None):
        phraseElement = SPhraseSpec(self)
        if verb is not None:
            # AG: fix here: check if "verb" is a VPPhraseSpec or a Verb
            if isinstance(verb, PhraseElement):
                phraseElement.setVerbPhrase(verb)
            else:
                phraseElement.setVerb(verb)
        if subject is not None:
            phraseElement.setSubject(subject)
        if directObject is not None:
            phraseElement.setObject(directObject)
        return phraseElement

    # Creates a canned text phrase with the given text.
    def createStringElement(self, text=None):
        return StringElement(text)

    # Creates a new coordinated phrase with two elements (initially)
    def createCoordinatedPhrase(self, coord1=None, coord2=None):
        return CoordinatedPhraseElement(coord1, coord2)

    #*********************************************************************************
    # Document level stuff
    #*********************************************************************************

    # Creates a new document element
    def createDocument(self, title=None, component=None):
        document = DocumentElement(DocumentCategory.DOCUMENT, title)
        if component is None:
            pass
        elif isinstance(component, list):
            document.addComponents(component)
        elif isinstance(component, NLGElement):
            document.addComponent(component)
        else:
            raise ValueError('Invalid component type: ' + str(type(component)))
        return document

    # Creates a new section element with the given title and adds the given
    # component.
    def createList(self, component=None):
        dlist = DocumentElement(DocumentCategory.LIST, None)
        if component is None:
            pass
        elif isinstance(component, list):
            dlist.addComponents(component)
        elif isinstance(component, NLGElement):
            dlist.addComponent(component)
        else:
            raise ValueError('Invalid component type: ' + str(type(component)))
        return dlist

    # Creates a new section element with the given title and adds the given
    # component.
    def createEnumeratedList(self, component=None):
        dlist = DocumentElement(DocumentCategory.ENUMERATED_LIST, None)
        if component is None:
            pass
        elif isinstance(component, list):
            dlist.addComponents(component)
        elif isinstance(component, NLGElement):
            dlist.addComponent(component)
        else:
            raise ValueError('Invalid component type: ' + str(type(component)))
        return dlist

    # Creates a list item for adding to a list element. The list item has the
    # given component.
    def createListItem(self, component=None):
        if component is None:
            return DocumentElement(DocumentCategory.LIST_ITEM, None)
        elif isinstance(component, NLGElement):
            listItem = DocumentElement(DocumentCategory.LIST_ITEM, None)
            listItem.addComponent(component)
            return listItem
        else:
            raise ValueError('Invalid component type: ' + str(type(component)))

    # Creates a new paragraph element and adds the given component
    def createParagraph(self, component=None):
        paragraph = DocumentElement(DocumentCategory.PARAGRAPH, None)
        if component is None:
            pass
        elif isinstance(component, list):
            paragraph.addComponents(component)
        elif isinstance(component, NLGElement):
            paragraph.addComponent(component)
        else:
            raise ValueError('Invalid component type: ' + str(type(component)))
        return paragraph

    # Creates a new section element.
    def createSection(self, title=None, component=None):
        if title is None and component is None:
            return DocumentElement(DocumentCategory.SECTION, None)
        if not isinstance(title, str):
            raise ValueError('Invalid title type: ' + str(type(title)))
        section = DocumentElement(DocumentCategory.SECTION, title)
        if component is None:
            pass
        elif isinstance(component, list):
            section.addComponents(component)
        elif isinstance(component, NLGElement):
            section.addComponent(component)
        else:
            raise ValueError('Invalid component type: ' + str(type(component)))
        return section

    # Creates a sentence
    def createSentence(self, a=None, b=None, c=None):
        sentence = DocumentElement(DocumentCategory.SENTENCE, None)
        if a is None and b is None and c is None:
            pass
        elif isinstance(a, list):
            sentence.addComponents(a)
        elif isinstance(a, NLGElement):
            sentence.addComponent(a)
        elif isinstance(a, str) and b is None and c is None:
            if a is not None:
                sentence.addComponent(self.createStringElement(a))
            return sentence
        else:
            sentence.addComponent(self.createClause(a, b, c))
        return sentence



from .ClauseHelper              import *
from .CoordinatedPhraseHelper   import *
from .NounPhraseHelper          import *
from .PhraseHelper              import *
from .SyntaxProcessor           import *
from .VerbPhraseHelper          import *



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from .NLGElement                    import *
from ..features.Feature             import *
from ..features.InternalFeature     import *
from ..features.NumberAgreement     import *
from .PhraseCategory                import *
from .StringElement                 import *


# This class defines coordination between two or more phrases.
class CoordinatedPhraseElement(NLGElement):
    PLURAL_COORDINATORS = ["and"]

    # Creates a blank coordinated phrase ready for new coordinates to be added.
    def __init__(self, coordinate1=None, coordinate2=None):
        super().__init__()
        if coordinate1 is None and coordinate2 is None:
            self.setFeature(Feature.CONJUNCTION, "and")
        else:
            self.addCoordinate(coordinate1)
            self.addCoordinate(coordinate2)
            self.setFeature(Feature.CONJUNCTION, "and")

    # Adds a new coordinate to this coordination.
    def addCoordinate(self, newCoordinate):
        coordinates = self.getFeatureAsElementList(InternalFeature.COORDINATES)
        if not coordinates:
            coordinates = []
            self.setFeature(InternalFeature.COORDINATES, coordinates)
        if isinstance(newCoordinate, NLGElement):
            if newCoordinate.isA(PhraseCategory.CLAUSE) and len(coordinates)>0:
                newCoordinate.setFeature(Feature.SUPRESSED_COMPLEMENTISER, True)
            coordinates.append(newCoordinate)
        elif isinstance(newCoordinate, str):
            coordElement = StringElement(newCoordinate)
            coordElement.setFeature(Feature.SUPRESSED_COMPLEMENTISER, True)
            coordinates.append(coordElement)
        self.setFeature(InternalFeature.COORDINATES, coordinates)

    # @Override
    def getChildren(self):
        return self.getFeatureAsElementList(InternalFeature.COORDINATES)

    # Clears the existing coordinates in this coordination. It performs exactly
    # the same as removeFeature(Feature.COORDINATES).
    def clearCoordinates(self):
        self.removeFeature(InternalFeature.COORDINATES)

    # Adds a new pre-modifier to the phrase element. Pre-modifiers will be
    # realised in the syntax before the coordinates.
    def addPreModifier(self, newPreModifier):
        preModifiers = self.getFeatureAsElementList(InternalFeature.PREMODIFIERS)
        if preModifiers is None:
            preModifiers = []
        if isinstance(newPreModifier, NLGElement):
            preModifiers.append(newPreModifier)
        elif isinstance(newPreModifier, str):
            preModifiers.append(StringElement(newPreModifier))
        else:
            raise ValueError('Invalid newPreModifier type: ' + str(type(newPreModifier)))
        self.setFeature(InternalFeature.PREMODIFIERS, preModifiers)

    # Retrieves the list of pre-modifiers currently associated with this
    # coordination.
    def getPreModifiers(self):
        return self.getFeatureAsElementList(InternalFeature.PREMODIFIERS)

    # Retrieves the list of complements currently associated with this
    # coordination.
    def getComplements(self):
        return self.getFeatureAsElementList(InternalFeature.COMPLEMENTS)

    # Adds a new post-modifier to the phrase element. Post-modifiers will be
    # realised in the syntax after the coordinates.
    def addPostModifier(self, newPostModifier):
        postModifiers = self.getFeatureAsElementList(InternalFeature.POSTMODIFIERS)
        if postModifiers is None:
            postModifiers = []
        if isinstance(newPostModifier, NLGElement):
            postModifiers.append(newPostModifier)
        elif isinstance(newPostModifier, str):
            postModifiers.append(StringElement(newPostModifier))
        else:
            raise ValueError('Invalid newPostModifier type: ' + str(type(newPostModifier)))
        self.setFeature(InternalFeature.POSTMODIFIERS, postModifiers)

    # Retrieves the list of post-modifiers currently associated with this
    # coordination.
    def getPostModifiers(self):
        return self.getFeatureAsElementList(InternalFeature.POSTMODIFIERS)

    # @Override
    def printTree(self, indent=None):
        if indent is None:
            indent = ''
        thisIndent      = indent + " |-"
        childIndent     = indent + " | "
        lastIndent      = indent + " \\-"
        lastChildIndent = indent + "   "
        pstr = "CoordinatedPhraseElement:\n"
        children = self.getChildren()
        for i, child in enumerate(children):
            if i < len(children)-1:
                pstr += thisIndent + child.printTree(childIndent)
            else:
                pstr += lastIndent + child.printTree(lastChildIndent)
        return pstr

    # Adds a new complement to the phrase element.
    def addComplement(self, newComplement):
        complements = self.getFeatureAsElementList(InternalFeature.COMPLEMENTS)
        if complements is None:
            complements = []
        if isinstance(newComplement, NLGElement):
            complements.append(newComplement)
        elif isinstance(newComplement, str):
            complements.append(StringElement(newComplement))
        else:
            raise ValueError('Invalid newComplement type: ' + str(type(newComplement)))
        self.setFeature(InternalFeature.COMPLEMENTS, complements)

    # A convenience method for retrieving the last coordinate in this
    # coordination.
    def getLastCoordinate(self):
        children = self.getChildren()
        if not children:
            return None
        return children[-1]

    # set the conjunction to be used in a coordinatedphraseelement
    # @param conjunction
    def setConjunction(self, conjunction):
        self.setFeature(Feature.CONJUNCTION, conjunction)

    # @return  conjunction used in coordinatedPhraseElement
    def getConjunction(self):
        return self.getFeatureAsString(Feature.CONJUNCTION)

    # @return true if this coordinate is plural in a syntactic sense
    def checkIfPlural(self):
        # doing this right is quite complex, take simple approach for now
        if len(self.getChildren()) == 1:
            return (NumberAgreement.PLURAL == self.getLastCoordinate().getFeature(Feature.NUMBER))
        else:
            return self.getConjunction() in self.PLURAL_COORDINATORS



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from copy import deepcopy
from ...features.DiscourseFunction      import *
from ...features.Feature                import *
from ...features.Gender                 import *
from ...features.InternalFeature        import *
from ...features.LexicalFeature         import *
from ...features.Person                 import *
from ...framework.InflectedWordElement  import *
from ...framework.LexicalCategory       import *
from ...framework.ListElement           import *
from ...framework.NLGElement            import *
from ...framework.PhraseCategory        import *
from ...framework.PhraseElement         import *
from ...framework.WordElement           import *
from .PhraseHelper                      import *


# This class contains static methods to help the syntax processor realise noun
# phrases.
class NounPhraseHelper(object):
    QUALITATIVE_POSITION = 1
    COLOUR_POSITION      = 2
    CLASSIFYING_POSITION = 3
    NOUN_POSITION        = 4

    # The main method for realising noun phrases.
    @classmethod
    def realise(cls, parent, phrase):
        realisedElement = ListElement()
        if phrase is not None and not phrase.getFeatureAsBoolean(Feature.ELIDED):
            realisedElement = ListElement()
            if phrase.getFeatureAsBoolean(Feature.PRONOMINAL):
                realisedElement.addComponent(cls.createPronoun(parent, phrase))
            else:
                cls.realiseSpecifier(phrase, parent, realisedElement)
                cls.realisePreModifiers(phrase, parent, realisedElement)
                cls.realiseHeadNoun(phrase, parent, realisedElement)
                PhraseHelper.realiseList(parent, realisedElement, \
                        phrase.getFeatureAsElementList(InternalFeature.COMPLEMENTS), \
                        DiscourseFunction.COMPLEMENT)
                PhraseHelper.realiseList(parent, realisedElement, phrase.getPostModifiers(), \
                        DiscourseFunction.POST_MODIFIER)
        return realisedElement

    # Realises the head noun of the noun phrase.
    @classmethod
    def realiseHeadNoun(cls, phrase, parent, realisedElement):
        headElement = phrase.getHead()
        if headElement is not None:
            headElement.setFeature(Feature.ELIDED, phrase.getFeature(Feature.ELIDED))
            headElement.setFeature(LexicalFeature.GENDER, phrase.getFeature(LexicalFeature.GENDER))
            headElement.setFeature(InternalFeature.ACRONYM, phrase.getFeature(InternalFeature.ACRONYM))
            headElement.setFeature(Feature.NUMBER, phrase.getFeature(Feature.NUMBER))
            headElement.setFeature(Feature.PERSON, phrase.getFeature(Feature.PERSON))
            headElement.setFeature(Feature.POSSESSIVE, phrase.getFeature(Feature.POSSESSIVE))
            headElement.setFeature(Feature.PASSIVE, phrase.getFeature(Feature.PASSIVE))
            currentElement = parent.realise(headElement)
            currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.SUBJECT)
            realisedElement.addComponent(currentElement)

    # Realises the pre-modifiers of the noun phrase. Before being realised,
    # pre-modifiers undergo some basic sorting based on adjective ordering.
    @classmethod
    def realisePreModifiers(cls, phrase,  parent, realisedElement):
        preModifiers = phrase.getPreModifiers()
        if phrase.getFeatureAsBoolean(Feature.ADJECTIVE_ORDERING):
            preModifiers = cls.sortNPPreModifiers(preModifiers)
        PhraseHelper.realiseList(parent, realisedElement, preModifiers, DiscourseFunction.PRE_MODIFIER)

    # Realises the specifier of the noun phrase.
    @classmethod
    def realiseSpecifier(cls, phrase, parent, realisedElement):
        specifierElement = phrase.getFeatureAsElement(InternalFeature.SPECIFIER)
        if specifierElement is not None and not phrase.getFeatureAsBoolean(InternalFeature.RAISED) and not \
                phrase.getFeatureAsBoolean(Feature.ELIDED):
            if not specifierElement.isA(LexicalCategory.PRONOUN) and \
                    specifierElement.getCategory() != PhraseCategory.NOUN_PHRASE:
                specifierElement.setFeature(Feature.NUMBER, phrase.getFeature(Feature.NUMBER))
            currentElement = parent.realise(specifierElement)
            if currentElement is not None:
                currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.SPECIFIER)
                realisedElement.addComponent(currentElement)

    # Sort the list of premodifiers for this noun phrase using adjective
    # ordering (ie, "big" comes before "red")
    @classmethod
    def sortNPPreModifiers(cls, originalModifiers):
        if originalModifiers is None or len(originalModifiers)<=1:
            orderedModifiers = originalModifiers
        else:
            orderedModifiers = deepcopy(originalModifiers)
            changesMade = True
            while changesMade:
                changesMade = False
                for i in range(len(orderedModifiers)-1):
                    if cls.getMinPos(orderedModifiers[i]) > cls.getMaxPos(orderedModifiers[i+1]):
                        temp = orderedModifiers[i]
                        orderedModifiers[i] = orderedModifiers[i+1]
                        orderedModifiers[i+1] = temp
                        changesMade = True
        return orderedModifiers


    # Determines the minimim position at which this modifier can occur.
    @classmethod
    def getMinPos(cls, modifier):
        position = cls.QUALITATIVE_POSITION
        if modifier.isA(LexicalCategory.NOUN) or modifier.isA(PhraseCategory.NOUN_PHRASE):
            position = cls.NOUN_POSITION
        elif modifier.isA(LexicalCategory.ADJECTIVE) or modifier.isA(PhraseCategory.ADJECTIVE_PHRASE):
            adjective = cls.getHeadWordElement(modifier)
            if adjective.getFeatureAsBoolean(LexicalFeature.QUALITATIVE):
                position = cls.QUALITATIVE_POSITION
            elif adjective.getFeatureAsBoolean(LexicalFeature.COLOUR):
                position = cls.COLOUR_POSITION
            elif adjective.getFeatureAsBoolean(LexicalFeature.CLASSIFYING):
                position = cls.CLASSIFYING_POSITION
        return position

    # Determines the maximim position at which this modifier can occur.
    @classmethod
    def getMaxPos(cls, modifier):
        position = cls.NOUN_POSITION
        if modifier.isA(LexicalCategory.ADJECTIVE) or modifier.isA(PhraseCategory.ADJECTIVE_PHRASE):
            adjective = cls.getHeadWordElement(modifier)
            if adjective.getFeatureAsBoolean(LexicalFeature.CLASSIFYING):
                position = cls.CLASSIFYING_POSITION
            elif adjective.getFeatureAsBoolean(LexicalFeature.COLOUR):
                position = cls.COLOUR_POSITION
            elif adjective.getFeatureAsBoolean(LexicalFeature.QUALITATIVE):
                position = cls.QUALITATIVE_POSITION
            else:
                position = cls.CLASSIFYING_POSITION
        return position

    # Retrieves the correct representation of the word from the element.
    @classmethod
    def getHeadWordElement(cls, element):
        head = None
        if isinstance(element, WordElement):
            head =  element
        elif isinstance(element, InflectedWordElement):
            head = element.getFeature(InternalFeature.BASE_WORD)
        elif isinstance(element, PhraseElement):
            head = cls.getHeadWordElement(element.getHead())
        return head

    # Creates the appropriate pronoun if the subject of the noun phrase is
    # pronominal.
    @classmethod
    def createPronoun(cls, parent, phrase):
        pronoun = "it"
        phraseFactory = phrase.getFactory()
        personValue = phrase.getFeature(Feature.PERSON)
        if Person.FIRST == personValue:
            pronoun = "I"
        elif Person.SECOND == personValue:
            pronoun = "you"
        else:
            genderValue = phrase.getFeature(LexicalFeature.GENDER)
            if Gender.FEMININE == genderValue:
                pronoun = "she"
            elif Gender.MASCULINE == genderValue:
                pronoun = "he"
        # AG: createWord now returns WordElement; so we embed it in an
        # inflected word element here
        proElement = phraseFactory.createWord(pronoun, LexicalCategory.PRONOUN)
        if isinstance(proElement, WordElement):
            element = InflectedWordElement(proElement)
            element.setFeature(LexicalFeature.GENDER, proElement.getFeature(LexicalFeature.GENDER))
            element.setFeature(Feature.PERSON, proElement.getFeature(Feature.PERSON))
        else:
            element = proElement
        element.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.SPECIFIER)
        element.setFeature(Feature.POSSESSIVE, phrase.getFeature(Feature.POSSESSIVE))
        element.setFeature(Feature.NUMBER, phrase.getFeature(Feature.NUMBER))
        if phrase.hasFeature(InternalFeature.DISCOURSE_FUNCTION):
            element.setFeature(InternalFeature.DISCOURSE_FUNCTION, \
                phrase.getFeature(InternalFeature.DISCOURSE_FUNCTION))
        return element



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ..features.DiscourseFunction       import *
from ..features.Feature                 import *
from ..features.Gender                  import *
from ..features.InternalFeature         import *
from ..features.LexicalFeature          import *
from ..features.Person                  import *
from ..framework.InflectedWordElement   import *
from ..framework.LexicalCategory        import *
from ..framework.NLGElement             import *
from ..framework.PhraseCategory         import *
from ..framework.PhraseElement          import *
from ..framework.WordElement            import *
from .AdjPhraseSpec                     import *


# This class defines a noun phrase. 
class NPPhraseSpec(PhraseElement):
    def __init__(self, phraseFactory):
        super().__init__(PhraseCategory.NOUN_PHRASE)
        self.setFactory(phraseFactory)

    # @see simplenlg.framework.PhraseElement#setHead(java.lang.Object) This
    # version sets NP default features from the head
    # @Override
    def setHead(self, newHead):
        super().setHead(newHead)
        self.setNounPhraseFeatures(self.getFeatureAsElement(InternalFeature.HEAD))

    # A helper method to set the features required for noun phrases, from the
    # head noun
    def setNounPhraseFeatures(self, nounElement):
        if nounElement is None:
            return
        self.setFeature(Feature.POSSESSIVE, nounElement.getFeatureAsBoolean(Feature.POSSESSIVE))
        self.setFeature(InternalFeature.RAISED, False)
        self.setFeature(InternalFeature.ACRONYM, False)
        if nounElement.hasFeature(Feature.NUMBER):
            self.setFeature(Feature.NUMBER, nounElement.getFeature(Feature.NUMBER))
        else:
            self.setPlural(False)
        if nounElement.hasFeature(Feature.PERSON):
            self.setFeature(Feature.PERSON, nounElement.getFeature(Feature.PERSON))
        else:
            self.setFeature(Feature.PERSON, Person.THIRD)
        if nounElement.hasFeature(LexicalFeature.GENDER):
            self.setFeature(LexicalFeature.GENDER, nounElement.getFeature(LexicalFeature.GENDER))
        else:
            self.setFeature(LexicalFeature.GENDER, Gender.NEUTER)
        if nounElement.hasFeature(LexicalFeature.EXPLETIVE_SUBJECT):
            self.setFeature(LexicalFeature.EXPLETIVE_SUBJECT, nounElement.getFeature(LexicalFeature.EXPLETIVE_SUBJECT))
        self.setFeature(Feature.ADJECTIVE_ORDERING, True)

    # sets the noun (head) of a noun phrase
    def setNoun(self, noun):
        nounElement = self.getFactory().createNLGElement(noun, LexicalCategory.NOUN)
        self.setHead(nounElement)

    # @return noun (head) of noun phrase
    def getNoun(self):
        return self.getHead()

    # setDeterminer - Convenience method for when a person tries to set
    #                 a determiner (e.g. "the") to a NPPhraseSpec.
    def setDeterminer(self, determiner):
        self.setSpecifier(determiner)

    # getDeterminer - Convenience method for when a person tries to get a
    #                 determiner (e.g. "the") from a NPPhraseSpec.
    def getDeterminer(self):
        return self.getSpecifier()

    # sets the specifier of a noun phrase. Can be determiner (eg "the"),
    # possessive (eg, "John's")
    def setSpecifier(self, specifier):
        if isinstance(specifier, NLGElement):
            self.setFeature(InternalFeature.SPECIFIER, specifier)
            specifier.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.SPECIFIER)
        else:
            # create specifier as word (assume determiner)
            specifierElement = self.getFactory().createWord(specifier, LexicalCategory.DETERMINER)
            # set specifier feature
            if specifierElement is not None:
                self.setFeature(InternalFeature.SPECIFIER, specifierElement)
                specifierElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.SPECIFIER)

    # @return specifier (eg, determiner) of noun phrase
    def getSpecifier(self):
        return self.getFeatureAsElement(InternalFeature.SPECIFIER)

    # Add a modifier to an NP Use heuristics to decide where it goes
    # @Override
    def addModifier(self, modifier):
        # string which is one lexicographic word is looked up in lexicon,
        # adjective is preModifier
        # Everything else is postModifier
        if modifier is None:
            return
        # get modifier as NLGElement if possible
        modifierElement = None
        if isinstance(modifier, NLGElement):
            modifierElement = modifier
        elif isinstance(modifier, str):
            modifierString = modifier
            if len(modifierString) > 0 and not " " in modifierString:
                modifierElement = self.getFactory().createWord(modifier, LexicalCategory.ANY)
        # if no modifier element, must be a complex string, add as postModifier
        if modifierElement is None:
            self.addPostModifier(modifier)
            return
        # AdjP is premodifer
        if isinstance(modifierElement, AdjPhraseSpec):
            self.addPreModifier(modifierElement)
            return
        # else extract WordElement if modifier is a single word
        modifierWord = None
        if modifierElement is not None and isinstance(modifierElement, WordElement):
            modifierWord = modifierElement
        elif modifierElement is not None and isinstance(modifierElement, InflectedWordElement):
            modifierWord = modifierElement.getBaseWord()
        # check if modifier is an adjective
        if modifierWord is not None and modifierWord.getCategory() == LexicalCategory.ADJECTIVE:
            self.addPreModifier(modifierWord)
            return
        # default case
        self.addPostModifier(modifierElement)



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ...features.DiscourseFunction          import *
from ...features.Feature                    import *
from ...features.InternalFeature            import *
from ...features.LexicalFeature             import *
from ...framework.CoordinatedPhraseElement  import *
from ...framework.InflectedWordElement      import *
from ...framework.LexicalCategory           import *
from ...framework.ListElement               import *
from ...framework.PhraseCategory            import *
from ...framework.WordElement               import *
from .PhraseHelper                          import *


# This class contains static methods to help the syntax processor realise
# coordinated phrases.
class CoordinatedPhraseHelper(object):
    # The main method for realising coordinated phrases.
    @classmethod
    def realise(cls, parent, phrase):
        realisedElement = None
        if phrase is not None:
            realisedElement = ListElement()
            PhraseHelper.realiseList(parent, realisedElement, phrase.getPreModifiers(), \
                    DiscourseFunction.PRE_MODIFIER)
            coordinated = CoordinatedPhraseElement()
            children = phrase.getChildren()
            conjunction = phrase.getFeatureAsString(Feature.CONJUNCTION)
            coordinated.setFeature(Feature.CONJUNCTION, conjunction)
            coordinated.setFeature(Feature.CONJUNCTION_TYPE, phrase.getFeature(Feature.CONJUNCTION_TYPE))
            conjunctionElement = None
            if children:
                if phrase.getFeatureAsBoolean(Feature.RAISE_SPECIFIER):
                    cls.raiseSpecifier(children)
                child = phrase.getLastCoordinate()
                child.setFeature(Feature.POSSESSIVE, phrase.getFeature(Feature.POSSESSIVE))
                child = children[0]
                cls.setChildFeatures(phrase, child)
                coordinated.addCoordinate(parent.realise(child))
                for index, child in enumerate(children):
                    if 0==index: continue
                    cls.setChildFeatures(phrase, child)
                    if phrase.getFeatureAsBoolean(Feature.AGGREGATE_AUXILIARY):
                        child.setFeature(InternalFeature.REALISE_AUXILIARY, False)
                    if child.isA(PhraseCategory.CLAUSE):
                        child.setFeature(Feature.SUPRESSED_COMPLEMENTISER, \
                                phrase.getFeature(Feature.SUPRESSED_COMPLEMENTISER))
                    #skip conjunction if it's null or empty string
                    if conjunction:
                        conjunctionElement = InflectedWordElement(conjunction, LexicalCategory.CONJUNCTION)
                        conjunctionElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.CONJUNCTION)
                        coordinated.addCoordinate(conjunctionElement)
                    coordinated.addCoordinate(parent.realise(child))
                realisedElement.addComponent(coordinated)
            PhraseHelper.realiseList(parent, realisedElement, phrase.getPostModifiers(), DiscourseFunction.POST_MODIFIER)
            PhraseHelper.realiseList(parent, realisedElement, phrase.getComplements(), DiscourseFunction.COMPLEMENT)
        return realisedElement

    # Sets the common features from the phrase to the child element.
    @classmethod
    def setChildFeatures(cls, phrase, child):
        if phrase.hasFeature(Feature.PROGRESSIVE):
            child.setFeature(Feature.PROGRESSIVE, phrase.getFeature(Feature.PROGRESSIVE))
        if phrase.hasFeature(Feature.PERFECT):
            child.setFeature(Feature.PERFECT, phrase.getFeature(Feature.PERFECT))
        if phrase.hasFeature(InternalFeature.SPECIFIER):
            child.setFeature(InternalFeature.SPECIFIER, phrase.getFeature(InternalFeature.SPECIFIER))
        if phrase.hasFeature(LexicalFeature.GENDER):
            child.setFeature(LexicalFeature.GENDER, phrase.getFeature(LexicalFeature.GENDER))
        if phrase.hasFeature(Feature.NUMBER):
            child.setFeature(Feature.NUMBER, phrase.getFeature(Feature.NUMBER))
        if phrase.hasFeature(Feature.TENSE):
            child.setFeature(Feature.TENSE, phrase.getFeature(Feature.TENSE))
        if phrase.hasFeature(Feature.PERSON):
            child.setFeature(Feature.PERSON, phrase.getFeature(Feature.PERSON))
        if phrase.hasFeature(Feature.NEGATED):
            child.setFeature(Feature.NEGATED, phrase.getFeature(Feature.NEGATED))
        if phrase.hasFeature(Feature.MODAL):
            child.setFeature(Feature.MODAL, phrase.getFeature(Feature.MODAL))
        if phrase.hasFeature(InternalFeature.DISCOURSE_FUNCTION):
            child.setFeature(InternalFeature.DISCOURSE_FUNCTION, phrase.getFeature(InternalFeature.DISCOURSE_FUNCTION))
        if phrase.hasFeature(Feature.FORM):
            child.setFeature(Feature.FORM, phrase.getFeature(Feature.FORM))
        if phrase.hasFeature(InternalFeature.CLAUSE_STATUS):
            child.setFeature(InternalFeature.CLAUSE_STATUS, phrase.getFeature(InternalFeature.CLAUSE_STATUS))
        if phrase.hasFeature(Feature.INTERROGATIVE_TYPE):
            child.setFeature(InternalFeature.IGNORE_MODAL, True)

    # Checks to see if the specifier can be raised and then raises it. In order
    # to be raised the specifier must be the same on all coordinates.
    @classmethod
    def raiseSpecifier(cls, children):
        allMatch = True
        child = None if not children else children[0]
        specifier = None
        test = None
        if child is not None:
            specifier = child.getFeatureAsElement(InternalFeature.SPECIFIER)
            if specifier is not None:
                # AG: this assumes the specifier is an InflectedWordElement or phrase.
                # it could be a Wordelement, in which case, we want the baseform
                if isinstance(specifier, WordElement):
                    test = specifier.getBaseForm()
                else:
                    specifier.getFeatureAsString(LexicalFeature.BASE_FORM)
            if test is not None:
                index = 1
                while index < len(children) and allMatch:
                    child = children[index]
                    if child is None:
                        allMatch = False
                    else:
                        specifier = child.getFeatureAsElement(InternalFeature.SPECIFIER)
                        if isinstance(specifier, WordElement):
                            childForm = specifier.getBaseForm()
                        else:
                            specifier.getFeatureAsString(LexicalFeature.BASE_FORM)
                        if not test == childForm:
                            allMatch = False
                    index += 1
                if allMatch:
                    for eachChild in range(1,len(children)):
                        child = children[eachChild]
                        child.setFeature(InternalFeature.RAISED, True)



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

import re

# This class is used to parse numbers that are passed as figures, to determine
# whether they should take "a" or "an" as determiner.
class DeterminerAgrHelper(object):
    # An array of strings which are exceptions to the rule that "an" comes
    # before vowels
    AN_EXCEPTIONS = [ r"one", r"180", r"110" ]
    # Start of string involving vowels, for use of "an"
    AN_AGREEMENT = r"\A(a|e|i|o|u).*"

    # Check whether this string starts with a number that needs "an" (e.g.
    # "an 18% increase")
    @classmethod
    def requiresAn(cls, string):
        req = False
        lowercaseInput = string.lower()
        if re.fullmatch(cls.AN_AGREEMENT,lowercaseInput) and not cls.isAnException(lowercaseInput):
            req = True
        else:
            numPref = cls.getNumericPrefix(lowercaseInput)
            if numPref and re.fullmatch(r"^(8|11|18).*$", numPref):
                num = int(numPref)
                req = cls.checkNum(num)
        return req

    # check whether a string beginning with a vowel is an exception and doesn't
    # take "an" (e.g. "a one percent change")
    @classmethod
    def isAnException(cls, string):
        for ex in cls.AN_EXCEPTIONS:
            if re.fullmatch(r"^" + ex + r".*", string):
                return True
        return False

    # Returns true if the number starts with 8, 11 or 18 and is
    # either less than 100 or greater than 1000, but excluding 180,000 etc.
    @classmethod
    def checkNum(cls, num):
        needsAn = False
        # eight, eleven, eighty and eighteen
        if num in [11, 18, 8] or (num >= 80 and num < 90):
            needsAn = True
        elif num > 1000:
            num = int(round(num / 1000))
            needsAn = cls.checkNum(num)
        return needsAn

    # Retrieve the numeral prefix of a string.
    @classmethod
    def getNumericPrefix(cls, instr):
        numeric = ''
        if instr is not None:
            instr = instr.strip()
            if instr:
                first = instr[0]
                if first.isdigit():
                    numeric += first
                    for next_loc in instr[1:]:
                        if next_loc.isdigit():
                            numeric += next_loc
                        # skip commas within numbers
                        elif next_loc == ',':
                            continue
                        else:
                            break
        if numeric:
            return str(numeric)
        return None

    # Check to see if a string ends with the indefinite article "a" and it agrees with {@code np}.
    @classmethod
    def checkEndsWithIndefiniteArticle(cls, text, np):
        tokens = text.split(" ")
        lastToken = tokens[-1]
        if lastToken.lower()=="a" and cls.requiresAn(np):
            tokens = tokens[:-1]
            tokens.append("an")
            return cls.stringArrayToString(tokens)
        return text

    # Turns ["a","b","c"] into "a b c"
    @classmethod
    def stringArrayToString(cls, sArray):
        buf = ''
        for i, char in enumerate(sArray):
            buf += char
            if i != len(sArray)-1:
                buf += ' '
        return buf



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='simplenlg',
    version='0.2.0',
    author='Brad Jascob',
    author_email='bjascob@msn.com',
    description='Python Implementation of SimpleNLG',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/bjascob/pySimpleNLG',
    include_package_data=True,
    package_data={'':['default-lexicon.xml']},
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Mozilla Public License 1.1 (MPL 1.1)',
        "Operating System :: OS Independent",
    ],
)



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

import re
from ...features.DiscourseFunction          import *
from ...features.Feature                    import *
from ...features.InternalFeature            import *
from ...framework.CoordinatedPhraseElement  import *
from ...framework.DocumentCategory          import *
from ...framework.DocumentElement           import *
from ...framework.ListElement               import *
from ...framework.NLGElement                import *
from ...framework.NLGModule                 import *
from ...framework.StringElement             import *


# This processing module deals with punctuation when applied to
# DocumentElements.
class OrthographyProcessor(NLGModule):
    def __init__(self):
        super().__init__()
        self.commaSepPremodifiers = None # set whether to separate premodifiers using commas
        self.commaSepCuephrase    = None # set whether to include a comma after a cue

    # @Override
    def initialise(self):
        self.commaSepPremodifiers = True
        self.commaSepCuephrase    = False

    # Check whether this processor separates premodifiers using a comma.
    def isCommaSepPremodifiers(self):
        return self.commaSepPremodifiers

    # Set whether to separate premodifiers using a comma. I
    def setCommaSepPremodifiers(self, commaSepPremodifiers):
        self.commaSepPremodifiers = commaSepPremodifiers

    # Check whether this processor separates cue phrases from a matrix phrase
    # using a comma.
    def isCommaSepCuephrase(self):
        return self.commaSepCuephrase

    # If set to true, separates a cue phrase from the matrix
    # phrase using a comma.
    def setCommaSepCuephrase(self, commaSepCuephrase):
        self.commaSepCuephrase = commaSepCuephrase

    # @Override
    def realise(self, element):
        if isinstance(element, NLGElement):
            return self._realiseElement(element)
        elif isinstance(element, list):
            return self._realiseElementList(element)
        else:
            raise ValueError('Invalid element type: ' + str(type(element)))

    def _realiseElement(self, element):
        realisedElement = None
        function = None #the element's discourse function
        #get the element's function first
        if isinstance(element, ListElement):
            children = element.getChildren()
            if children:
                firstChild = children[0]
                function = firstChild.getFeature(InternalFeature.DISCOURSE_FUNCTION)
        elif element is not None:
            function = element.getFeature(InternalFeature.DISCOURSE_FUNCTION)
        if element is not None:
            category = element.getCategory()
            if isinstance(category, DocumentCategory) and isinstance(element, DocumentElement):
                components = element.getComponents()
                if category == DocumentCategory.SENTENCE:
                    realisedElement = self.realiseSentence(components, element)
                if category == DocumentCategory.LIST_ITEM:
                    if components:
                        # recursively realise whatever is in the list item
                        # NB: this will realise embedded lists within list items
                        realisedElement = ListElement(self.realise(components))
                        realisedElement.setParent(element.getParent())
                else:
                    element.setComponents(self.realise(components))
                    realisedElement = element
            elif isinstance(element, ListElement):
                # AG: changes here: if we have a premodifier, then we ask the
                # realiseList method to separate with a comma.
                # if it's a postmod, we need commas at the start and end only
                # if it's appositive
                buffer = ''
                if DiscourseFunction.PRE_MODIFIER == function:
                    all_appositives = True
                    for child in element.getChildren():
                        all_appositives = all_appositives and child.getFeatureAsBoolean(Feature.APPOSITIVE)
                    if all_appositives:
                        buffer += ", "
                    sep = "," if self.commaSepPremodifiers else ''
                    buffer = self.realiseList(buffer, element.getChildren(), sep)
                    if all_appositives:
                        buffer += ", "
                elif DiscourseFunction.POST_MODIFIER == function:
                    postmods = element.getChildren()
                    # bug fix due to Owen Bennett
                    length = len(postmods)
                    for i in range(length):
                        postmod = postmods[i]
                        # if the postmod is appositive, it's sandwiched in commas
                        if postmod.getFeatureAsBoolean(Feature.APPOSITIVE):
                            buffer += ", "
                            buffer += str(self._realiseElement(postmod))
                            if i < length-1:
                                buffer += ", "
                        else:
                            buffer += str(self._realiseElement(postmod))
                            if isinstance(postmod, ListElement) or \
                               (postmod.getRealisation() is not None and not postmod.getRealisation()==""):
                                buffer  += " "
                elif (DiscourseFunction.CUE_PHRASE==function or DiscourseFunction.FRONT_MODIFIER==function) and \
                    self.commaSepCuephrase:
                    sep = ',' if self.commaSepCuephrase else ''
                    buffer = self.realiseList(buffer, element.getChildren(), sep)
                else:
                    buffer = self.realiseList(buffer, element.getChildren(), "")
                realisedElement = StringElement(buffer)
            elif isinstance(element, CoordinatedPhraseElement):
                realisedElement = self.realiseCoordinatedPhrase(element.getChildren())
            else:
                realisedElement = element
            # make the realised element inherit the original category
            # essential if list items are to be properly formatted later
            if realisedElement is not None:
                realisedElement.setCategory(category)
            #check if this is a cue phrase; if param is set, postfix a comma
            if (DiscourseFunction.CUE_PHRASE==function or DiscourseFunction.FRONT_MODIFIER==function) and \
                    self.commaSepCuephrase:
                realisation = realisedElement.getRealisation()
                if not realisation.endswith(","):
                    realisation = realisation + ","
                realisedElement.setRealisation(realisation)
        #remove preceding and trailing whitespace from internal punctuation
        self.removePunctSpace(realisedElement)
        return realisedElement

    # removes extra spaces preceding punctuation from a realised element
    def removePunctSpace(self, realisedElement):
        if realisedElement is not None:
            realisation = realisedElement.getRealisation()
            if realisation is not None:
                realisation = re.sub(" ,", ",",  realisation)
                realisation = re.sub(",,+", ",", realisation)
                realisedElement.setRealisation(realisation)

    # Performs the realisation on a sentence. This includes adding the
    # terminator and capitalising the first letter.
    def realiseSentence(self, components, element):
        realisedElement = None
        if components:
            realisation = ''
            realisation = self.realiseList(realisation, components, "")
            realisation = self.stripLeadingCommas(realisation)
            realisation = self.capitaliseFirstLetter(realisation)
            realisation = self.terminateSentence(realisation, element.getFeatureAsBoolean(InternalFeature.INTERROGATIVE))
            element.clearComponents()
            element.setRealisation(realisation)
            realisedElement = element
        return realisedElement

    # Adds the sentence terminator to the sentence. This is a period ('.') for
    # normal sentences or a question mark ('?') for interrogatives.
    def terminateSentence(self, realisation, interrogative):
        character = realisation[-1]
        if character != '.' and character != '?':
            if interrogative:
                realisation += '?'
            else:
                realisation += '.'
        return realisation

    # Remove recursively any leading spaces or commas at the start
    # of a sentence.
    def stripLeadingCommas(self, realisation):
        character = realisation[0]
        if character == ' ' or character == ',':
            realisation =  realisation[1:]
            realisation = self.stripLeadingCommas(realisation)
        return realisation

    # Capitalises the first character of a sentence if it is a lower case letter.
    def capitaliseFirstLetter(self, realisation):
        character = realisation[0]
        if character.islower():
            realisation = realisation[0].upper() + realisation[1:]
        return realisation

    # @Override
    def _realiseElementList(self, elements):
        realisedList = []
        if elements:
            for eachElement in elements:
                if isinstance(eachElement, DocumentElement):
                    realisedList.append(self.realise(eachElement))
                else:
                    realisedList.append(eachElement)
        return realisedList

    # Realises a list of elements appending the result to the on-going
    # realisation.
    def realiseList(self, realisation, components, listSeparator):
        realisedChild = None
        for i, thisElement in enumerate(components):
            realisedChild = self.realise(thisElement)
            childRealisation = realisedChild.getRealisation()
            # check that the child realisation is non-empty
            if childRealisation and not re.fullmatch(r"^[\s\n]+$", childRealisation):
                realisation += realisedChild.getRealisation()
                if len(components)>1 and i<len(components)-1:
                    realisation += listSeparator
                realisation += ' '
        if len(realisation) > 0:
            realisation = realisation[:-1]
        return realisation

    # Realises coordinated phrases.
    def realiseCoordinatedPhrase(self, components):
        realisation = ''
        realisedChild = None
        length = len(components)
        for index in range(length):
            realisedChild = components[index]
            if index < length - 2 and \
                    DiscourseFunction.CONJUNCTION==realisedChild.getFeature(InternalFeature.DISCOURSE_FUNCTION):
                realisation += ", "
            else:
                realisedChild = self.realise(realisedChild)
                realisation += realisedChild.getRealisation() + ' '
        realisation = realisation[:-1]
        realisation = realisation.replace(" ,", ",")
        return StringElement(realisation)



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from enum import Enum

# An enumeration representing the different forms a verb and its associated
# phrase can take.
class Form(Enum):
    # The bare infinitive is the base form of the verb.
    BARE_INFINITIVE = 0
    # In English, the gerund form refers to the usage of a verb as a noun. For
    # example, I like swimming. In more general terms, gerunds
    # are usually formed from the base word with -ing added to the
    # end.
    GERUND = 1
    # The imperative form of a verb is the one used when the grammatical
    # mood is one of expressing a command or giving a direct request.
    IMPERATIVE = 2
    # The infinitive form represents the base form of the verb, with our
    # without the particle "to".
    INFINITIVE = 3
    # Normal form represents the base verb.
    NORMAL = 4
    # Most verbs will have only a single form for the past tense. However, some
    # verbs will have two forms, one for the simple past tense and one for the
    # past participle (also knowns as passive participle or perfect
    # participle). The part participle represents the second of these two
    # forms.
    PAST_PARTICIPLE = 5
    # The present participle is identical in form to the gerund and is normally
    # used in the active voice. However, the gerund is meant to highlight a
    # verb being used as a noun. The present participle remains as a verb.
    PRESENT_PARTICIPLE = 6

    def __str__(self):
        return self.name
